#pragma once
#include ".\WinClass.h"

typedef unsigned char byte;

class SpotEmitter;	//forward decl, as all sub-emitters are spot emitters
class Emitter;		//forward decl, for ParticleCollision warp

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~PARTICLE CLASS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Particle
{
public:
	mathVector position;//Particle's postion
	mathVector velocity;//Particle's velocity
	long age;			//time since it was emitted

	//constructor
	Particle(float xPos,float yPos, float xVel, float yVel);
};


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~SPACE WARPS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//A space warp creates an acceleration, a deflection etc...
//ie: anything that changes the velocity of a particle...

//This is the Space Warp Base Class.
//It's pointer is used to store derived class pointers
class SpaceWarp
{
	char ID[64];//string ID. used when a spacewarp is deleted from an
				//emitter. Each warp SHOULD have different ID string.
protected:
	bool deflector;
public:
	SpaceWarp(const char * id);
	bool isDeflector();
	friend Particle;
	char * getID();
	void setID(const char *);

	virtual int applyWarp(Particle& p)=0;
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~Field Space Warp~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//It creates an acceleration along one direction vector.

class SW_Field : public SpaceWarp
{
protected:
	mathVector acceleration;//direction along which field acceleration is applied
public:
	SW_Field(char* id,float xAcc=0,float yAcc=0);
	float& xAcceleration();
	float& yAcceleration();
	int applyWarp(Particle& p);
};

//creates a localised field (in a circular region) along one direction vector.
class SW_LocalField : public SW_Field
{
	mathVector position;
	int decayOrder;
	float maxDistance,minDistance;
public:
	SW_LocalField(char* id,float xPosition,float yPosition,
					float xAcc,float yAcc,int attenuation);
	float& X();
	float& Y();
	int& Decay();
	float& MaxInfluence();
	float& MinInfluence();
	int applyWarp(Particle& p);
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~Attractor Space Warp~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//It creates an acceleration towards one point.
class SW_PointAttractor : public SpaceWarp
{
	float acc;

	mathVector multiplier;
	mathVector position;
	int decayOrder;
	float maxDistance,minDistance;

public:
	SW_PointAttractor(char* id,float xPos=0, float yPos=0,
						float acc=0,int attenuation=0);
	int applyWarp(Particle& p);
	float& Acceleration();
	float& XMultiplier();
	float& YMultiplier();
	float& X();
	float& Y();
	int& Decay();
	float& MaxInfluence();
	float& MinInfluence();
};


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~SpeedMultiplier Space Warp~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//causes particles to change speed, depending on current speed.

class SW_SpeedMultiplier : public SpaceWarp
{
	float multiplier;	//particle velocity = velocity * multiplier
						//if value is -ve, go crazy.
	float maxSpeed;
public:
	SW_SpeedMultiplier(char* id, float value);
	float& Multiplier();
	float& SpeedLimit();
	int applyWarp(Particle& p);
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~Deflector Line Space Warp~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//It acts like a wall.

class SW_Deflector_Line : public SpaceWarp
{
	float bounce;
    mathVector point1;
	mathVector point2;
	mathVector unitVect;

	float slope;//in equation y=mx+c, slope=m, c=kvalue
	bool vertical;//if slope = infinity
	float kvalue;

	//check if collision with this deflector occurs when moving from pt1 to pt2
	//and return true if collision occurs, also fill collidePt with point of collision.
	bool getCollision(mathVector pt1,mathVector pt2,mathVector * collidePt );

public:
	SW_Deflector_Line(char* id, float x1,float y1,float x2, float y2, float bounce=1);
	float &X1();
	float &Y1();
	float &X2();
	float &Y2();
	float &BounceCoeff();
	void getSlope();

	int applyWarp(Particle& p);
};

class SW_ParticleForces : public SpaceWarp
{
	Emitter* pSystem;
	SW_PointAttractor* pWarp;
	int skipStep;	//forces are exerted by every skipStep'th
					//particle each time.
	int currentStep;//stores current step. always 0 if skipStep=1

public:
	SW_ParticleForces(char* id, Emitter * e, SW_PointAttractor* w);
	int& SkipRate();
	void setEmitter(Emitter* e);
	Emitter* getEmitter();
	void setPWarp(SW_PointAttractor* w);
	SW_PointAttractor* getPWarp();
	int applyWarp(Particle& p);
};

class SW_ParticleCollisions : public SpaceWarp
{
	float bounce;
	Emitter* pSystem;
	int skipStep;	//collisions occur only for every skipStep'th
					//particle each time.
	int currentStep;//stores current step. always 0 if skipStep=1
	float thresh;
public:
	SW_ParticleCollisions(char* id, Emitter * e, float bounce);
	int& SkipRate();
	float& BounceCoeff();
	float& Threshold();
	void setEmitter(Emitter* e);
	Emitter* getEmitter();
	int applyWarp(Particle& p);
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~EMITTERS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This is the emitter base class.
//It contains the general definition of an emitter
//with particles, associated space warps...

class Emitter : public ImageClass
{
	char ID[64];
	byte r0,g0,b0;		//start rgb color.
	byte r1,g1,b1;		//end rgb color.
	bool blend;			//if particles are blended with background

protected:
	Particle* p;		//This has an array of particles.
	SpaceWarp **sw;		//This contains an array of pointers to SpaceWarps.

	SpotEmitter * subEmitter;	//A pointer to the sub-emitter. Each particle will
								//act as an emitter now. 

	long deflectorStart;//Stores index of first deflector in sw[i];

	long maxCollisions;	//Maximum times that a collision iteration occurs per particle.

	long maxWarps;
	long numWarps;
	long maxParticles;
	long numParticles;
	bool initialized;
	long lifetime;
	bool hasImage;

public:
	Emitter(long numParticles=100,long numWarps=10);
	~Emitter();

	Particle * getPArray();
	long getParticleLimit();
	long getWarpLimit();
	void setParticleLimit(long numParticles);
	void setWarpLimit(long numWarps);
	char * getID();
	void setID(char *);

	void setInitColor(int,int,int);
	void setFinalColor(int,int,int);
	long& LifeTime();
	long& MaximumCollisions();
	bool& graphicEmitter();
	bool& AlphaBlend();

	long NumberOfParticles();
	long NumberOfWarps();

	bool addWarp(SpaceWarp * newWarp);
	bool delWarp(char * warpID);

	void drawParticles(HDC hdc);
	void delParticle(long index);
	void simulate();

	SpotEmitter * SubEmitter();
	void SubEmitter(SpotEmitter * subemit);

	virtual bool emitParticle()=0;

};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~SpotEmitter Class~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class SpotEmitter: public Emitter
{
	mathVector position;
	long rate;			//number of particles emitted per cycle.
	float initialSpeed;
	float speedVariation;
	float startAngle;
	float endAngle;

public:
	SpotEmitter(char* id,long numParticles=100,long numWarps=10);
    float& X();
	float& Y();
	float& pSpeed();
	float& pSpeedVar();
	long& EmitRate();
	void setEmitAngle(float,float);
	bool emitParticle();
};

class BoxEmitter: public Emitter
{
	mathVector bound1,bound2;
	long rate;			//number of particles emitted per cycle.

public:
	BoxEmitter(char* id,long numParticles=100,long numWarps=10);
    float& X1();
	float& Y1();

	float& X2();
	float& Y2();
	long& EmitRate();
	bool emitParticle();
};
